@@ -16,6 +16,7 @@ DEFAULT_KEY_NAME = 'kommandos-key'
 
 DEFAULT_DNS_TTL = 86400
 
+S3_BUCKET_NAME = "kommandos-bucket"
 KOMMANDOS_HOME_FOLDER = f'{os.path.expanduser("~")}/.aws-kommandos'
 
 if 'win' in sys.platform:
@@ -79,8 +80,8 @@ def get_arguments():
                                    help="Specify this flag if you want to poll the SSH service on remote "
                                         "after the server boot. "
                                         "The netcat tool needs to be installed on the system to use that. "
-                                        "Keep in mind that you should provide this flag in case you wanna "
-                                        "trigger script invocation on remote, "
+                                        "Keep in mind that you should provide this flag in case you want to "
+                                        "enable script invocation on remote, "
                                         "since Kommandos needs to know when the service is available for invoking it. "
                                         "Also make sure an inbound rule in the corresponding security group "
                                         "for connecting to the 22/tcp port.")
@@ -131,6 +132,20 @@ def get_arguments():
                                    help="Specify the instance name. "
                                         f"Default is '{DEFAULT_INSTANCE_NAME}'.")
 
+    connect = parser.add_argument_group("EC2 Connection Commands")
+    connect.add_argument('--connect',
+                         dest='connect',
+                         required=False,
+                         help='Specify an instance-id to connect to. '
+                              'The SSH port (22/tcp) should be reachable for doing that. '
+                              "The private key is taken from the argument --key-pair-name and "
+                              "is fetched from the Kommandos S3 bucket if doesn't exist locally.")
+    connect.add_argument('--user',
+                         dest='user',
+                         required=False,
+                         help='Specify a username to use while connecting to the instance via SSH. '
+                              'If omitted, the default user name of the AMI image is used.')
+
     # Route53 Managing Domains
     route53 = parser.add_argument_group('Route53 Domain Management')
     route53.add_argument('--add-record',
@@ -286,7 +301,8 @@ def get_arguments():
                           help="Specify the key pair name to use alongside the created instance. "
                                f"Default is '{DEFAULT_KEY_NAME}'. "
                                f"The keys managed by Kommandos "
-                               f"are stored under the '{KOMMANDOS_HOME_FOLDER}' directory.")
+                               f"are stored under the '{KOMMANDOS_HOME_FOLDER}' directory "
+                               f"and are uploaded to the Kommandos managed S3 bucket.")
     ssh_keys.add_argument("--force-recreate-key",
                           dest="force_recreate_key",
                           action='store_true',
@@ -410,6 +426,7 @@ class AwsManager:
                  aws_access_key_secret: str = None,
                  region_name: str = None):
         self.home_folder = KOMMANDOS_HOME_FOLDER
+        self.s3_bucket_name = S3_BUCKET_NAME
         try:
             self.ec2 = boto3.resource('ec2', aws_access_key_id=aws_access_key_id,
                                       aws_secret_access_key=aws_access_key_secret,
@@ -420,15 +437,106 @@ class AwsManager:
             self.ec2_client = boto3.client('ec2', aws_access_key_id=aws_access_key_id,
                                            aws_secret_access_key=aws_access_key_secret,
                                            region_name=region_name)
-
+            self.s3_client = boto3.client('s3', aws_access_key_id=aws_access_key_id,
+                                          aws_secret_access_key=aws_access_key_secret,
+                                          region_name=region_name)
             if not os.path.exists(self.home_folder):
                 os.makedirs(self.home_folder)
                 print(f"The Kommandos's home folder has been created at '{self.home_folder}'. "
-                      f"The SSH private keys that are to be created by Kommandos will be stored there.")
+                      f"The SSH private keys to be created by Kommandos will be stored there.")
+
+            if not self.get_bucket(self.s3_bucket_name):
+                print("Creating a new Kommandos S3 bucket, as it doesn't seem to exist yet")
+                self.create_bucket(bucket_name=self.s3_bucket_name)
+
         except Exception as e:
             print(f"Kommandos initialization failed: {e}")
             exit(1)
 
+    ## S3 BUCKETS
+    ### GETTING ALL BUCKETS
+    def get_all_buckets(self):
+        return self.s3_client.list_buckets()['Buckets']
+
+    ### GETTING BUCKET BY NAME
+    def get_bucket(self, bucket_name: str):
+        buckets = self.get_all_buckets()
+        for bucket in buckets:
+            for k, v in bucket.items():
+                if 'Name' == k and bucket_name == v:
+                    return bucket
+
+    ### CREATING A BUCKET
+    def create_bucket(self, bucket_name: str):
+        print(f"Creating a new private S3 bucket with name '{bucket_name}'")
+        try:
+            self.s3_client.create_bucket(
+                ACL='private',
+                Bucket=bucket_name,
+                CreateBucketConfiguration={
+                    'LocationConstraint': self.s3_client.meta.region_name
+                }
+            )
+        except Exception as e:
+            if "you already own it" in f"{e}":
+                print(f"The S3 bucket '{bucket_name}' already exists")
+            else:
+                print(f"Failed to create a S3 bucket: {e}")
+
+    ### DELETING A BUCKET
+    def delete_bucket(self, bucket_name: str):
+        print(f"Deleting the S3 bucket: '{bucket_name}'")
+        try:
+            self.s3_client.delete_bucket(
+                Bucket=bucket_name
+            )
+        except Exception as e:
+            if "The specified bucket does not exist" in f"{e}":
+                print(f"The S3 bucket '{bucket_name}' doesn't exist")
+            else:
+                print(f"Failed to delete the S3 bucket: {e}")
+
+    ### GETTING A FILE FROM S3
+    def download_file_from_bucket(self, bucket_name, remote_file_name, local_file_name):
+        print(f"Downloading '{remote_file_name}' from s3://{bucket_name} to '{local_file_name}'")
+        try:
+            self.s3_client.download_file(
+                Bucket=bucket_name,
+                Key=remote_file_name,
+                Filename=local_file_name
+            )
+            if os.path.exists(local_file_name):
+                print(f"{remote_file_name} has been downloaded")
+                return True
+        except Exception as e:
+            if "Not Found" in f"{e}":
+                print(f"The file '{remote_file_name}' has not been found on '{bucket_name}' bucket")
+            else:
+                print(f"Failed to download the file from the S3 bucket: {e}")
+
+    ### UPLOADING A FILE TO S3
+    def upload_file_to_bucket(self, bucket_name, local_file_name):
+        print(f"Uploading '{local_file_name}' to s3://{bucket_name}")
+        try:
+            self.s3_client.upload_file(
+                Bucket=bucket_name,
+                Filename=local_file_name,
+                Key=os.path.basename(local_file_name)
+            )
+        except Exception as e:
+            print(f"Failed to upload a file from to the S3 bucket: {e}")
+
+    ### DELETING A FILE FROM S3
+    def delete_file_from_bucket(self, bucket_name, remote_file_name):
+        print(f"Deleting '{remote_file_name}' from s3://{bucket_name}")
+        try:
+            self.s3_client.delete_object(
+                Bucket=self.s3_bucket_name,
+                Key=remote_file_name
+            )
+        except Exception as e:
+            print(f"Failed to upload a file from to the S3 bucket: {e}")
+
     ## DNS
     ### GETTING ALL HOSTING ZONES
     def get_all_hosted_zones(self):
@@ -528,6 +636,15 @@ class AwsManager:
                 key_pairs.append(key_pair)
         return key_pairs
 
+    def download_key_pair_from_s3(self, key_pair_name: str):
+        key_path = f"{self.home_folder}/{key_pair_name}.pem"
+        self.download_file_from_bucket(
+            bucket_name=self.s3_bucket_name,
+            remote_file_name=f"{key_pair_name}.pem",
+            local_file_name=key_path
+        )
+        os.chmod(f"{self.home_folder}/{key_pair_name}.pem", 0o400)
+
     def print_key_pairs(self, verbose: bool = False):
         key_pairs = self.get_key_pairs()
         if key_pairs:
@@ -566,6 +683,9 @@ class AwsManager:
             # 0x400 -r--------
             os.chmod(key_path, 0o400)
 
+            self.upload_file_to_bucket(bucket_name=self.s3_bucket_name,
+                                       local_file_name=key_path)
+
     ### DELETE A KEY PAIR
     def delete_key_pair(self, key_name: str):
         print(colored(f"Deleting the SSH key pair: {key_name}", 'yellow'))
@@ -577,6 +697,8 @@ class AwsManager:
             os.chmod(key_path, 0o600)
             os.remove(key_path)
 
+        self.delete_file_from_bucket(self.s3_bucket_name, file_name)
+
     ## SECURITY GROUPS
     ### GET SECURITY GROUPS
     def get_all_security_groups(self):
@@ -944,12 +1066,14 @@ class AwsManager:
                 inst = {
                     'InstanceId': instance.id,
                     'PublicIpAddress': instance.public_ip_address,
-                    'KeyPairName': instance.key_pair.key_name,
                     'SecurityGroups': instance.security_groups
                 }
                 if name:
                     inst['Name'] = name
 
+                if instance.key_pair:
+                    inst['KeyPairName'] = instance.key_pair.key_name
+
                 if verbose:
                     inst['ImageId'] = instance.image_id
 
@@ -1020,13 +1144,6 @@ class AwsManager:
             instance = self.get_instance(new_instance['InstanceId'])
             instance.wait_until_running()
             print(f'The server is up and running at {colored(instance.public_ip_address, "green")}')
-            if poll_ssh:
-                print('Waiting until the SSH service is available...')
-                self.poll_ssh_status(instance.id)
-            identified_user_name = self.get_default_ami_user_name(instance.image_id)
-            key_path = f"{self.home_folder}/{key_pair_name}.pem"
-            print("Use the following command for connecting to the instance: " +
-                  colored(f"ssh {identified_user_name}@{instance.public_ip_address} -i {key_path}", 'green'))
             return instance
 
     def invoke_script(self, instance_id: str,
@@ -1108,6 +1225,7 @@ def main():
     instance_name = options.instance_name
 
     domain_name = options.fqdn
+    poll_ssh = options.poll_ssh
 
     aws_manager = AwsManager(aws_access_key_id=access_key_id,
                              aws_access_key_secret=access_key_secret,
@@ -1197,7 +1315,7 @@ def main():
             print(f"{len(images)} AMI images found")
         else:
             print('No images found')
-    elif options.start:
+    elif options.connect:
         if options.force_recreate_key:
             aws_manager.delete_key_pair(key_name=key_pair_name)
         try:
@@ -1209,6 +1327,42 @@ def main():
                 print(f"{type(e)} {e}")
                 exit(1)
 
+        instance_id = options.connect
+        print(f"Connecting to '{instance_id}'")
+        instance = aws_manager.get_instance(instance_id=instance_id)
+
+        if hasattr(instance, 'key_name') and instance.key_name:
+            if instance.key_name == key_pair_name:
+                key_path = f"{aws_manager.home_folder}/{key_pair_name}.pem"
+                if not os.path.exists(key_path):
+                    aws_manager.download_key_pair_from_s3(key_pair_name)
+
+                if options.user:
+                    user_name = options.user
+                else:
+                    user_name = aws_manager.get_default_ami_user_name(instance.image_id)
+                key_path = f"{aws_manager.home_folder}/{key_pair_name}.pem"
+                ssh_cmd = f"ssh {user_name}@{instance.public_ip_address} -i {key_path}"
+                print(colored(f"Using the following command: {ssh_cmd}", 'green'))
+                os.system(ssh_cmd)
+            else:
+                print(colored("The instance's key name doesn't match the default one or the one specified "
+                      "with the --key-pair-name flag", 'red'))
+        else:
+            print(colored("The instance doesn't seem to have an SSH key pair attached", 'red'))
+
+    elif options.start:
+        if options.force_recreate_key:
+            aws_manager.delete_key_pair(key_name=key_pair_name)
+        try:
+            aws_manager.create_key_pair(key_name=key_pair_name)
+        except Exception as e:
+            if 'InvalidKeyPair.Duplicate' in f"{e}":
+                print(colored(f"The SSH kehy pair with the name '{key_pair_name}' already exists", 'yellow'))
+            else:
+                print(f"{type(e)} {e}")
+                exit(1)
+
         new_instance = aws_manager.run_instance(image_id=image_id,
                                                 key_pair_name=key_pair_name,
                                                 security_group_id=security_group_id,
@@ -1227,9 +1381,27 @@ def main():
                                           record_value=f"1 {new_instance.public_ip_address}",
                                           ttl=ttl)
         if options.invoke_script:
+            if poll_ssh:
+                print('Waiting until the SSH service is available...')
+                aws_manager.poll_ssh_status(new_instance.id)
+
             aws_manager.invoke_script(instance_id=new_instance.instance_id,
                                       file_name=options.invoke_script,
                                       parameters=options.invoke_script_argument)
+        if poll_ssh:
+            print('Waiting until the SSH service is available...')
+            aws_manager.poll_ssh_status(new_instance.id)
+            print("Connecting to the created instance")
+
+            key_path = f"{aws_manager.home_folder}/{key_pair_name}.pem"
+            if os.path.exists(key_path):
+                print("The private key already exists on the file system, use --force-recreate-key to overwrite")
+            else:
+                aws_manager.download_key_pair_from_s3(key_pair_name)
+
+            identified_user_name = aws_manager.get_default_ami_user_name(new_instance.image_id)
+            key_path = f"{aws_manager.home_folder}/{key_pair_name}.pem"
+            os.system(f"ssh {identified_user_name}@{new_instance.public_ip_address} -i {key_path}")
 
 
 if __name__ == '__main__':
